@@ -38,25 +38,43 @@ class Experiment:
         else:
             self.n_workers = multiprocessing.cpu_count()
 
-        members = self.run()
-        group_theory_data = stabalize(members, cutoff, n_trials)
-        self.groups = [
-            Group(
-                member_data, self.survey, weights=group_theory_data[2]
-                ) for member_data in group_theory_data[0]]
-        group_dicts = [group.__dict__ for group in self.groups]
-        self.group_df = pd.DataFrame(group_dicts)
-        self.group_df.insert(0, 'group_id', np.arange(len(self.group_df)))
+        self.cutoff = cutoff
 
-        self._add_group_info_to_df()
-        self._drop_unnecessary_columns_from_group_df()
-        self.group_table = self._convert_pd_to_fitstable(self.group_df)
-        self.galaxy_table = self._convert_pd_to_fitstable(self.survey.data_frame)
-        self.edge_data = group_theory_data[1]
-
-    def run(self, use_multiprocessing: bool = True):
-        """Runs the algorithm."""
+    def run(self, 
+            use_multiprocessing: bool = True):
+        """
+        Runs the algorithm.
+        
+        Parameters
+        ----------
+        use_multiprocessing : bool
+            Whether to use multiprocessing for parallelisation of experiment trial runs. Default is true.
+        
+        Returns
+        -------
+            self.members : np.array
+                The list of members and associated galaxy group IDs to each member of a group
+        """
         if use_multiprocessing:
+            try:
+                # to ensure forwards and backwards compatibility with various OS versions which use spawn rather than fork 
+                # as the default process creation method. Spawn creates thread-related Runtime Errors related to children spawning 
+                # faster than parent processes can finish spawning, and require tighter controls not necessary for these use-cases.
+                multiprocessing.set_start_method('fork') 
+            except RuntimeError as runtime_error:
+                # Just print(e) is cleaner and more likely what we want,
+                # but if lets insist on printing message specifically whenever possible...
+                if hasattr(runtime_error, 'message'):
+                    error_message = runtime_error.message
+                else:
+                    error_message = runtime_error
+                
+                if str(error_message) == 'context has already been set':
+                    print('WARN: Multiprocessing context has already been set and is trying to be reset. You can only set this context once per Python session. Continuing with default')
+                else:
+                    print('ERROR: Multiprocessing has thrown an error on setting context to "fork". Traceback raised is printed below. Recommended setting is to run with use_multiprocessing=False and raise an issue in the GitHub repository for pyFoF.')
+                    raise
+            
             with rp.Progress(
                 "[progress.description]{task.description}",
                 rp.BarColumn(),
@@ -111,7 +129,7 @@ class Experiment:
                         members_list = [group.members for group in concatenated_results]
                         
             #assert len(results) == self.number_of_trials
-            return members_list
+            self.members = members_list
         else:
             results = [
                 Trial(
@@ -122,7 +140,31 @@ class Experiment:
                 ]
             concatenated_results = np.concatenate(results)
             members_list = [group.members for group in concatenated_results]
-            return members_list
+            self.members = members_list
+
+        # run stabilisation, i.e. graph theory post-processing
+        graph_theory_post_processed_output = stabalize(self.members, self.cutoff, self.number_of_trials)
+
+        # retrieve groups
+        self.groups = [
+            Group(
+                member_data, self.survey, weights=graph_theory_post_processed_output[2]
+                ) for member_data in graph_theory_post_processed_output[0]]
+        group_dicts = [group.__dict__ for group in self.groups]
+        # transform groups into Pandas DataFrame with group IDs as group_id
+        self.group_df = pd.DataFrame(group_dicts)
+        self.group_df.insert(0, 'group_id', np.arange(len(self.group_df)))
+
+        # clean up operations for above dataframe (not passed as parameter, used as class var)
+        self._add_group_info_to_df()
+        self._drop_unnecessary_columns_from_group_df()
+
+        # transform as FITS table to be exported for other secondary use-cases by user
+        self.group_table = self._convert_pd_to_fitstable(self.group_df)
+        self.galaxy_table = self._convert_pd_to_fitstable(self.survey.data_frame)
+        self.edge_data = graph_theory_post_processed_output[1]
+
+        return self.members
 
     def _add_group_info_to_df(self):
         """Adds the group ID and the galaxy weights to the survey data frame."""
@@ -141,10 +183,54 @@ class Experiment:
 
     @staticmethod
     def _convert_pd_to_fitstable(data_frame: pd.DataFrame) -> Table:
-        """Takes a pandas dataframe and returns a fits table."""
+        """
+        Takes a pandas dataframe and returns a fits table.
+        """
         fits_table = Table.from_pandas(data_frame)
         return fits_table
 
+    def get_survey_input_data(self):
+        """
+        Function to return the survey input data passed into the experiment
+        """
+        return self.survey.data_frame
+
+    def get_groups(self, output_type='pandas'):
+        """
+        Function to return the galaxy groups calculated by the FoF algorithm and experiments
+
+        Parameters
+        ----------
+        output_type : str
+            The output type to return, the options here are 'pandas' for a Pandas DataFrame or
+            'fits' for a FITS-friendly table.
+        
+        Returns
+        -------
+
+        """
+        if output_type not in ['pandas', 'fits']:
+            raise ValueError('Incorrect value specified for output_type argument. Accepted values are "pandas" and "fits".')
+        if output_type == 'pandas':
+            return self.group_df
+        elif output_type == 'fits':
+            return self.group_table
+
+    def get_edge_data(self):
+        """
+        Function to return the edge data calculated after the graph theory post-processing
+        """
+        return self.edge_data
+    
+    def write_edge_data(self) -> None:
+        """Writes the edge data as a txt file for edge visualizaton."""
+        f = open('edge_data.txt', 'w')
+        f.write('# ID1, ID2, WEIGHT')
+        edge_data = self.get_edge_data()
+        for edge_datum in edge_data:
+            f.write(f'{int(edge_datum[0])} {int(edge_datum[1])} {edge_datum[2]} \n')
+        f.close()
+        
     def write_group_catalog(self, outfile_name: str, overwrite = False) -> None:
         """Generates a group catalog as a fits file."""
         self.group_table.write(outfile_name, overwrite = overwrite)
@@ -160,7 +246,7 @@ class Experiment:
 
 if __name__ == '__main__':
     INFILE = './data/Kids/Kids_S_hemispec_no_dupes_updated.tbl'
-    
+
     #INFILE = './data/Kids/WISE-SGP_redshifts_w1mags.tbl'
     #INFILE = './data/Test_Data/Test_Cat.tbl'
     cosmo = FlatLambdaCDM(H0=70, Om0=0.3)
